/**************************************************************************
*                                                                         *
*                             COPYRIGHT NOTICE                            *
*                                                                         *
* This software/database is categorized as "United States Government      *
* Work" under the terms of the United States Copyright Act.  It was       *
* produced as part of the author's official duties as a Government        *
* employee and thus can not be copyrighted.  This software/database is    *
* freely available to the public for use without a copyright notice.      *
* Restrictions can not be placed on its present or future use.            *
*                                                                         *
* Although all reasonable efforts have been taken to ensure the accuracy  *
* and reliability of the software and data, the National Library of       *
* Medicine (NLM) and the U.S. Government do not and can not warrant the   *
* performance or results that may be obtained by using this software,     *
* data, or derivative works thereof.  The NLM and the U.S. Government     *
* disclaim any and all warranties, expressed or implied, as to the        *
* performance, merchantability or fitness for any particular purpose or   *
* use.                                                                    *
*                                                                         *
* In any work or product derived from this material, proper attribution   *
* of the author(s) as the source of the software or data would be         *
* appreciated.                                                            *
*                                                                         *
**************************************************************************/
/*
* BLASTX - Search a PROTEIN database for regions that approximately match
* regions of a TRANSLATED NUCLEOTIDE query sequence.  The basic command
* syntax is:
*
* 	BLASTX database query_sequence [parameters] [[top][bottom]]
*
* where:
* 	database names a file containing protein sequences, separated by
* 		header lines beginning with the character '>'.  A typical
* 		sequence entry is:
*
*		>CCHU (PIR) Cytochrome c - Human | 1.0,1.0,1.0,1.0,1.0
*		GDVEKGKKIFIMKCSQCHTVEKGGKHKTGP
*		NLHGLFGRKTGQAPGYSYTAANKNKGIIWG
*		EDTLMEYLENPKKYIPGTKMIFVGIKKKEE
*		RADLIAYLKKATNE
*
* 	query_sequence names a file containing a nucleotide sequence (where the
* 		header line is optional).
* 
* The BLASTX command permits optional additional arguments, such as X=50,
* that reset certain internal parameters.  The available parameters are:
*
*	E or e is the expected number of HSPs
*	S or s gives the High-scoring Segment Pair (HSP) cutoff score (default 45)
*	M or m is the name of a file containing substitution costs
*		(default is "BLOSUM62")
*	T or t gives the threshold for a word match (default is 17)
*	X or x gives the value for terminating word extensions (default is 20)
* Thus a typical command is
*
*	BLASTX AABANK SEQ E=50 M=PAM250
*
*	The "top" and "bottom" options must be specified last on the
*	command line, and direct BLASTX to translate only the top or
*	only the bottom strand, respectively.
*
*   (Note:  set tabstops every 4 columns for a better appearance).
*/
#include <ncbi.h>
#include <gish.h>
#include <gishlib.h>
#include <sys/stat.h>

#ifndef BLASTX
#define BLASTX
#endif

#define EXTERN
#include "blastapp.h"
#include "aabet.h"	/* the alphabet */
#include "ntbet.h"
#include "gcode.h"

#define DEFAULT_M	"BLOSUM62"	/* default scoring matrix */
#define DEFAULT_T	13	/* default threshold for neighborhood word hits */
#define DEFAULT_X	20	/* default score drop-off during hit extensions */
#define DEFAULT_W	3	/* default word size */
#define DEFAULT_E2	0.15
#define DEFAULT_S2	0

char
	*qname = "NO NAME",
	*qtitle = "NO NAME",
	*dbtitle,
	*dbbuf,			/* Memory-resident database */
	*s_seq, *ss_seq,
	*qq_seq,
	*t_seq,		/* temporary sequence storage */
	*n_seq,		/* nucleotide query sequence + null terminator */
	*q_seq,		/* query sequence + null terminator */
	*n_rev,		/* reverse complement of query sequence */
	M[FILENAME_MAX] = DEFAULT_M;	/* file containing the substitution scores */

int		qnamelen = 7, qtitlelen = 7;

int		n_len;		/* length of nucleotide query sequence */

long	totdblen;	/* total no. of letters in the database */

unsigned long
	maxlen,		/* maximum length of a database sequence */
	numwords,	/* number of critical words in hash table */
	numexact;	/* number of exact words that did not score at least T */

Score_t
	S2 = DEFAULT_S2,
	S = DEFAULT_S,		/* cutoff score for High-scoring Segment Pairs (HSP) */
	T = DEFAULT_T,		/* threshold for word hits */
	X = DEFAULT_X;		/* constant for word-hit extensions */

static int
	top=1,				/* TRUE ==> search top strand */
	bottom=1,			/* TRUE ==> search bottom strand */
	W = DEFAULT_W,		/* word size */
	C = DEFAULT_C,		/* genetic code */
	S_set = FALSE,
	S2_set = FALSE,
	T_set = FALSE,
	X_set = FALSE,
	W_set = FALSE,
	C_set = FALSE,
	E_set = FALSE,
	E2_set = TRUE,
	M_set = FALSE,
	cdi_flag = FALSE,
	Y_set = FALSE,
	Z_set = FALSE;

double
	E2 = DEFAULT_E2,
	E = DEFAULT_E;		/* Expected number of HSPs achieving cutoff score */

double
	Lambda_std, K_std, H_std;

static unsigned char	aatab[64], revaatab[64]; /* translation tables */

static char	xa_btoa[65];

double	ccode[64], cfq[64], cbits[64];

seq_data	sd[6] = {
	{0,(char *)0, (DIAG_TYPE *)0, (DIAG_TYPE *)0, (DIAG_PTR)0, (DIAG_PTR)0,
			0,0, 0,0,0,0,0,0, 0,0, 1, 0., 0.,0.,0.},
	{1,(char *)0, (DIAG_TYPE *)0, (DIAG_TYPE *)0, (DIAG_PTR)0, (DIAG_PTR)0,
			0,0, 0,0,0,0,0,0, 0,0, 2, 0., 0.,0.,0.},
	{2,(char *)0, (DIAG_TYPE *)0, (DIAG_TYPE *)0, (DIAG_PTR)0, (DIAG_PTR)0,
			0,0, 0,0,0,0,0,0, 0,0, 3, 0., 0.,0.,0.},
	{3,(char *)0, (DIAG_TYPE *)0, (DIAG_TYPE *)0, (DIAG_PTR)0, (DIAG_PTR)0,
			0,0, 0,0,0,0,0,0, 0,0, -1, 0., 0.,0.,0.},
	{4,(char *)0, (DIAG_TYPE *)0, (DIAG_TYPE *)0, (DIAG_PTR)0, (DIAG_PTR)0,
			0,0, 0,0,0,0,0,0, 0,0, -2, 0., 0.,0.,0.},
	{5,(char *)0, (DIAG_TYPE *)0, (DIAG_TYPE *)0, (DIAG_PTR)0, (DIAG_PTR)0,
			0,0, 0,0,0,0,0,0, 0,0, -3, 0., 0.,0.,0.}
		},	*sdp;


static int
	minframe=0,
	maxframe=6;

static MFILE
	*sfile,			/* file pointer for sequences */
	*tfile;			/* file pointer to table */

static CharPtr	cdi_file; /* codon frequencies filename */

static Score_t
	highs,		/* highest score in the PAM matrix */
	lows;		/* lowest score in the PAM matrix */

static unsigned long
	*seq_beg; 		/* starts of sequences in file or memory-resident dbase */

DFAPtr	dp;	/* the automaton */

double	Ecomp;

int		alphasize;			/* alpabet size */
int		minchar, maxchar;	/* min. and max. letter values */
static char
	/* array of amino acid names, plus null terminator */
	achars[ALPHASIZE_MAX+1],
	/* order[i][j] = j-th best char to sub. for i */
	order[ALPHAVAL_MAX+1][ALPHASIZE_MAX];

Score_t
	/* max cost of substituting for a character */
	maxcost[ALPHAVAL_MAX+1],
	/* matrx[i][j] = cost of replacing i by j */
	matrx[ALPHAVAL_MAX+1][ALPHAVAL_MAX+1];

Nlm_Boolean	consistency_flag = TRUE;

static void	search_x PROTO((MPPROTO));
static void	initdb PROTO((char *dbname));
static int	get_seq PROTO((MPPROTOC unsigned long));
static void	dosearch PROTO((Nlm_VoidPtr userp, TaskBlkPtr tp));
static int	print_x PROTO((HitListPtr));

unsigned long	dfanstates, dfastatesize, dfasize, dfaextent;

Cputime_t	t0, t1, t2, t3;

char	*citation = "Gish, Warren and David J. States (1993).  Identification of protein coding regions by database similarity search.  Nature Genetics 3:266-72.\nAltschul, Stephen F., Warren Gish, Webb Miller, Eugene W. Myers, and David J. Lipman (1990).  Basic local alignment search tool.  J. Mol. Biol. 215:403-410.";
#if 0
char	*citation = "Altschul, Stephen F., Warren Gish, Webb Miller, Eugene W. Myers, and David J. Lipman (1990).  Basic local alignment search tool.  J. Mol. Biol. 215:403-410.";
#endif

char	*notice = "statistical significance is estimated under the assumption that the equivalent of one reading frame in the query sequence codes for protein and that significant alignments will involve only coding reading frames.";

/*
Version	Date		Description
-------	-------		------------------------------------------------------------
1.1.7	12-13-90	Fixed hit extension bugs (improper use of parameter X and
					incorrect rigorous searches)
1.1.8	12-14-90	Using more common code, new hit lists
1.1.9	1-14-91		Fixed bug in bldaa.c regarding perfect match words
1.1.10	1-15-91		Fixed bug in Poisson P-value calculation
1.1.11	1-16-91		Fixed another bug in Poisson P-value calculation
1.1.12	1-28-91		Display one-line description of each db sequence hit
1.1.13	2-4-91		Word-wrap sequence descriptions
1.1.14	3-12-91		Added calculation of default T for the case when W==3
1.1.15	22-7-91		Default X is calculated using information theory
1.1.17	24-8-91		Proper translation of ambiguous nucleotides;
1.1.18	25-8-91		Fixed bug in extend_x()--diag_top was being updated twice!
                    This would cause data to become corrupted when many word
					hits were obtained, as diag_stack was effectively half the
					size it should have been.
1.1.19	13-9-91		Changed behavior of parameters V and B.
1.1.20	25-9-91		Improved reporting of individual HSP statistics
1.2.0	1-10-91		Improved Poisson statistics, screening and sorting of hits
1.2.1	23-10-91	Fixed bug in print_parms()
1.2.2	23-12-91	Improvement in command line parsing, new -overlap option
1.2.3	26-12-91	Improved error reporting; new K & L options
1.2.4	6-1-92	Only use unambiguous letters (non-X) when calculating K & L
1.2.5	18-2-92	Switched to new dfa library
1.2.6	9-3-92	Faster K calculation in karlin()
1.2.7	31-3-92	Added gap character '-' to the amino acid alphabet
1.2.8	12-6-92	Fixed bug in lambdak() when smaller alphabets are used
1.2.9	16-6-92	Added several hitlist sorting options
1.2.10	18-6-92 Corrected the observed high score reported
1.2.11	25-8-92	DEC Ultrix compatibility
1.2.12	4-9-92	Fixed bug in lib/hsppool.c
1.2.13	17-11-92	Added filter option
1.3.0	20-11-92	Added consistency to Poisson Event counts
1.3.1   9-2-93	Added codon usage option
1.3.2	17-2-93	Added low-cutoff (E2/S2) support
1.3.3   23-4-93 Added consistent Poisson P-value calculation
1.3.4   5-5-93  Karlin H was assumed to have wrong units in lib/stolen.c
1.3.5	9-6-93	Added gapdecayrate suggested by Phil Green
1.3.6	21-6-93	Fixed bug in consistent N count calculation for (-) strand hits
1.3.7   22-6-93 Fixed bug in consistp.c implementation of R(i,3) -- Phil Green
1.3.8   25-6-93 Finished gapdecayrate -- modified cutoffs()
1.3.9	7-7-93	Added -overlap2 option, prompted by Erik Sonnhammer
1.3.10	7-9-93	Cleaned up some of the E2/S2 usage.
1.3.11	23-9-93	Made E2=0.15 the default
1.3.12	29-10-93	Added -dbrecmin and -dbrecmax options
1.3.13	30-6-94	Updated the genetic codes
*/
CharPtr	version = "1.3.13";
CharPtr	rel_date = "30-Jun-94";


int
main(argc, argv)
	int		argc;
	char	**argv;
{
	char	o = NULLB, *a;
	CharPtr	theArg;
	size_t	nheaders;
	int	ac, i, j, arglen;
	Nlm_Boolean	secondary;
	double	x;
	long	ilong;

	sys_times(&t0);
	alarmprocset();

	banner(argv[0], version, rel_date, citation, notice);

	if (argc < 3)
		usage();

	for (ac=3; ac < argc; ++ac) {
		arglen = strlen(theArg = argv[ac]);
		if (str_casecmp("-over2", theArg) == 0 || str_casecmp("-overlap2", theArg) == 0) {
			ovlap = ovlap_x2;
			overlap_option = theArg;
			continue;
		}
		if (str_ncasecmp(theArg, "-over", 5) == 0) {
			ovlap = ovlap_0;
			overlap_option = theArg;
			continue;
		}
		if (str_ncasecmp(theArg, "-filter", 7) == 0) {
			if (++ac >= argc)
				bfatal(ERR_INVAL, "Missing filter specification");
			filtercmd = pick_filter(argv[ac], AMINO_ACID_SEQTYPE);
			continue;
		}
		if (str_ncasecmp("-consis", theArg, 7) == 0) {
			consistency_flag = FALSE;
			continue;
		}
		if (str_ncasecmp(theArg, "-poisson", 8) == 0) {
			continue;
		}
		if (str_casecmp(theArg, "-gi") == 0) {
			gi_option = TRUE;
			continue;
		}
		if (arglen >= 6 && arglen <= 11 &&
				str_ncasecmp("-echofilter", theArg, arglen) == 0) {
			echofilter_flag = TRUE;
			continue;
		}
		if (str_casecmp(theArg, "-dbrecmax") == 0) {
			if (++ac >= argc)
				bfatal(ERR_INVAL, "Missing -dbrecmax value");
			if (sscanf(argv[ac], "%ld", &dbrecmax) != 1 || dbrecmax <= 0)
				bfatal(ERR_INVAL, "Invalid -dbrecmax value:  %s", argv[ac]);
			continue;
		}
		if (str_casecmp(theArg, "-dbrecmin") == 0) {
			if (++ac >= argc)
				bfatal(ERR_INVAL, "Missing -dbrecmin value");
			if (sscanf(argv[ac], "%ld", &dbrecmin) != 1 || dbrecmin <= 0)
				bfatal(ERR_INVAL, "Invalid -dbrecmin value:  %s", argv[ac]);
			--dbrecmin;
			continue;
		}
		if (str_casecmp(theArg, "top") == 0) {
			bottom=0;
			continue;
		}
		if (str_casecmp(theArg, "bottom") == 0) {
			top=0;
			continue;
		}
		if (str_ncasecmp("-qoffset", theArg, 5) == 0) {
			if (++ac >= argc)
				bfatal(ERR_INVAL, "Missing -qoffset");
			if (sscanf(argv[ac], "%ld", &ilong) != 1)
				bfatal(ERR_INVAL, "Invalid value specified with -qoffset:  %s", argv[ac]);
			Qoffset = ilong;
			continue;
		}
		if (str_casecmp("-sort_by_pvalue", theArg) == 0) {
			hitlist_cmp_func = cmp_by_pval;
			continue;
		}
		if (str_casecmp("-sort_by_count", theArg) == 0) {
			hitlist_cmp_func = cmp_by_count;
			continue;
		}
		if (str_casecmp("-sort_by_highscore", theArg) == 0) {
			hitlist_cmp_func = cmp_by_highscore;
			continue;
		}
		if (str_casecmp("-sort_by_totalscore", theArg) == 0) {
			hitlist_cmp_func = cmp_by_totalscore;
			continue;
		}
		if (str_ncasecmp("-gapdecayrate", theArg, 8) == 0) {
			if (++ac >= argc)
				bfatal(ERR_INVAL, "Missing gapdecayrate");
			if (sscanf(argv[ac], "%lg", &gapdecayrate) != 1)
				bfatal(ERR_INVAL, "Invalid value specified with -gapdecayrate:  %s", argv[ac]);
			continue;
		}
		if (str_casecmp("-codoninfo", theArg) == 0) {
			if (++ac >= argc)
				bfatal(ERR_INVAL, "Filename argument missing for -codoninfo option");
			cdi_file = str_dup(argv[ac]);
			cdi_flag = TRUE;
			continue;
		}
		secondary = FALSE;
		if (theArg[1] == '=')
			o = theArg[0];
		else
			if (theArg[0] == '-')
				o = theArg[1];
			else
				if (theArg[1] == '2' && theArg[2] == '=') {
					o = theArg[0];
					secondary = TRUE;
				}
				else
					bfatal(ERR_INVAL, "Argument %d (``%s'') was unexpected or has improper form.", ac, theArg);
		a = theArg + 2;
		if (*a == NULLB) /* ignore */
			continue;
		if (secondary)
			++a;
		if (islower(o))
			o = toupper(o);
		if (!secondary)
		switch (o) {
			case 'P':
#ifdef MPROC_AVAIL
				numprocs = atoi(a);
				if (numprocs <= 0)
					bfatal(ERR_INVAL, "Bad value for P (the number of processors)");
				if (numprocs > THREADS_MAX)
					warning("%s:  Pmax = %d", module, THREADS_MAX);
				numprocs = MIN(numprocs, THREADS_MAX);
#else
				warning("Parameter P is ignored--this program has not been compiled to use multiprocessing", module);
#endif
				break;
#if 0
			case 'A':
				if ((int)strlen(a) >= FILENAME_MAX)
					bfatal(ERR_INVAL, "Codon frequencies filename is too long");
				A = str_dup(a);
				A_set = TRUE;
				break;
#endif
			case 'M':
				if ((int)strlen(a) >= FILENAME_MAX)
					bfatal(ERR_INVAL, "PAM file name is too long");
				strcpy(M, a);
				if (strcmp(M, DEFAULT_M) != 0)
					M_set = TRUE;
				break;
			case 'W':
				if (sscanf(a, "%d", &W) != 1 || W <= 0)
					bfatal(ERR_INVAL, "Bad value for W");
				if (W > WORDSIZE_MAX)
					bfatal(ERR_WORDSIZE, "W is too big--must be less than %d", WORDSIZE_MAX+1);
				W_set = TRUE;
				break;
			case 'S':
				if ((S = atol(a)) < 1)
					bfatal(ERR_INVAL, "Bad value for S");
				S_set = TRUE;
				break;
			case 'E':
				if (sscanf(a, "%lg", &E) != 1 || E <= 0. || E > DEFAULT_E_MAX)
					bfatal(ERR_INVAL, "Bad value for E");
				E_set = TRUE;
				break;
			case 'T':
				if (sscanf(a, "%ld", &ilong) != 1 || (T = ilong) < 1)
					bfatal(ERR_INVAL, "Bad value for T");
				T_set = TRUE;
				break;
			case 'X':
				if (sscanf(a, "%ld", &ilong) != 1 || (X = ilong) < 1 || X > 1000000)
					bfatal(ERR_INVAL, "Bad value for X");
				X_set = TRUE;
				break;
			case 'Y':
				if (sscanf(a, "%lg", &x) != 1 || x < 0.5 || x > LONG_MAX)
					bfatal(ERR_INVAL, "Bad value for Y");
				Meff = x + 0.5;
				Y_set = TRUE;
				break;
			case 'Z':
				if (sscanf(a, "%lg", &x) != 1 || x < 0.5 || x > LONG_MAX)
					bfatal(ERR_INVAL, "Bad value for Z");
				Neff = x + 0.5;
				Z_set = TRUE;
				break;
			case 'C':
				if (isalpha(*a)) {
					display_gcodes();
					exit_code(ERR_NONE);
				}
				C = atoi(a);
				if (find_gcode(C) == NULL) {
					printf("Bad value for C\n");
					display_gcodes();
					exit_code(ERR_NONE);
				}
				C_set = TRUE;
				break;
			case 'K':
				if (sscanf(a, "%lg", &K) != 1 || K < 0.)
					bfatal(ERR_INVAL, "Bad value for K");
				K_set = TRUE;
				break;
			case 'L':
				if (sscanf(a, "%lg", &Lambda) != 1 || Lambda <= 0.)
					bfatal(ERR_INVAL, "Bad value for L(ambda)");
				L_set = TRUE;
				break;
			case 'V':
				if (sscanf(a, "%ld", &ilong) != 1 || (V = ilong) < 0)
					bfatal(ERR_INVAL, "Bad value for V");
				V_set = TRUE;
				break;
			case 'B':
				if (sscanf(a, "%ld", &ilong) != 1 || (showblast = ilong) < 0)
					bfatal(ERR_INVAL, "Bad value for B");
				break;
			case 'H':
				showhist = atoi(a);
				break;
			default:
				usage();
		}
		else
		/* Secondary search options */
		switch (o) {
			case 'E':
				if (sscanf(a, "%lg", &E2) != 1 || E2 > 1000.0 || E2 < 0.)
					bfatal(ERR_INVAL, "Bad value for E2, the secondary expect value.");
				E2_set = TRUE;
				if (E2 == 0.)
					E2_set = FALSE;
				break;
			case 'S':
				if ((S2 = atol(a)) < 1 || S2 > 1000000)
					bfatal(ERR_INVAL, "Bad value for S2, the secondary cutoff score.");
				S2_set = TRUE;
				break;
			default:
				usage();
		}
	}
	if (top == 0 && bottom == 0)
		top = bottom = 1;

	minframe = top ? 0 : 3;
	maxframe = bottom ? 6 : 3;

	if (showblast == 0 && showhist == 0 && V == 0)
		bfatal(ERR_INVAL, "All output is suppressed--no search will be performed");

	if (cdi_flag)
		get_codonfq(cdi_file, ccode, cfq, cbits);

	init_gcode(find_gcode(C), aatab, revaatab);
	for (i = 0; i < 64; ++i)
		xa_btoa[i+1] = aa_btoa[aatab[i]];

	/* read file of substitution weights */
	if (!cdi_flag)
		get_weights(M, AAID_MAX+1, aa_atob, &alphasize, &minchar, &maxchar,
					&achars, &order, &lows, &highs, &maxcost, &matrx);
	else
		get_weightx(M, aatab, cbits, AAID_MAX+1, aa_atob, &alphasize, &minchar, &maxchar,
					&achars, &order, &lows, &highs, &maxcost, &matrx);

	get_query(argv[2]);
	wrap(stdout, "Query=  ", qtitle, qtitlelen, 79, qnamelen+1);
	printf("        (%s letters", lu2s(n_len));
	Qlen = n_len + Qoffset;
	if (Qoffset != 0)
		printf(", offset %s)", ld2s(Qoffset));
	printf(")\n\n");

	(void) printf("  Translating ");
	if (top && bottom)
		(void) printf("both strands of query sequence in all 6 reading frames\n");
	if (top && !bottom)
		(void) printf("top strand of query sequence in 3 reading frames\n");
	if (!top && bottom)
		(void) printf("bottom strand of query sequence in 3 reading frames\n");

	xlat_query();

	putchar('\n');
	initdb(argv[1]);
	(void) fflush(stdout);

	if (!Y_set)
		Meff = n_len;

	init_diags(); /* must come after the call to initdb() so dbsize is known */

	lambdak_std(lows, highs, &Lambda_std, &K_std, &H_std);

	if (cdi_flag) {
		if (top) {
			lambdak0();
			lambdakcnt(n_seq, n_len);
			lambdakcodon(lows, highs, &sd[0].efflen, &sd[0].Lambda, &sd[0].K, &sd[0].H);
			for (frame = 1; frame < 3; ++frame) {
				sd[frame].efflen = sd[0].efflen;
				sd[frame].Lambda = sd[0].Lambda;
				sd[frame].K = sd[0].K;
				sd[frame].H = sd[0].H;
			}
		}
		if (bottom) {
			lambdak0();
			lambdakcnt(n_rev, n_len);
			lambdakcodon(lows, highs, &sd[3].efflen, &sd[3].Lambda, &sd[3].K, &sd[3].H);
			for (frame = 4; frame < 6; ++frame) {
				sd[frame].efflen = sd[3].efflen;
				sd[frame].Lambda = sd[3].Lambda;
				sd[frame].K = sd[3].K;
				sd[frame].H = sd[3].H;
			}
		}
	}

	/*
	It might be better to call lambdak() for every sequence in
	the database, but then the program would be much slower.
	Another suggestion is to call lambdak() before reporting
	the probability of an HSP.  But for speed, lambdak() is
	called only once.
	*/
	i = 0;
	for (frame = minframe; frame < maxframe; ++frame) {
		sdp = &sd[frame];
		if (!cdi_flag) {
			sdp->K = K;
			sdp->Lambda = Lambda;
			if (lambdak(lows, highs, sdp->seq, sdp->len, &sdp->efflen,
					&sdp->Lambda, &sdp->K, &sdp->H) != 0) {
				warning("Could not calculate Lambda, K, and H in frame %d",
							sdp->stdframe);
				sdp->T = 0;
				sdp->S = 1;
				sdp->E = 0.;
				++i;
				continue;
			}
		}
		if (Y_set)
			sdp->efflen = Meff/CODON_LEN;
			
		sdp->T = T;
		if (T_set == FALSE) {
			if (W == 3) {
				/* Ad hoc equation for PAM matrices with scale = ln(2)/2 */
				sdp->T = ceil((5.076 + 1.3*log(sdp->H))/sdp->Lambda - 0.5);
				/*sdp->T = ceil((2.0 + 2.741*sdp->H)/sdp->Lambda - 0.5);*/
			}
			if (W == 4) {
				/* Ad hoc equation for PAM matrices with scale = ln(2)/2 */
				sdp->T = ceil((2.8 + 3.751*sdp->H)/sdp->Lambda - 0.5);
			}
		}
		/* Determine the cutoff score, S */
		if (S2_set || E2_set) {
			Score_t	s, hs;

			if (S2_set && !E2_set)
				sdp->E = 0.;
			else
				sdp->E = E2;
			sdp->S = S2;
			cutoffs(&sdp->S, &sdp->E, &hs, sdp->Lambda, sdp->K, sdp->H,
						300, 300, gapdecayrate);
			x = E;
			s = S;
			cutoffs(&s, &x, &sdp->highs_exp, sdp->Lambda, sdp->K, sdp->H, Neff, sdp->efflen, gapdecayrate);
			if (S_set)
				Ecomp += x;
			else
				Ecomp += E;

			/* Bet on how significant the least-significant hit will be */
			if (sdp->E * Neff / 300. > 100.)
				sdp->S0 = MAX(1, MIN(sdp->S, etos(sdp->E*Neff/30., sdp->Lambda, sdp->K, sdp->H, Neff, sdp->efflen)));
			else
				sdp->S0 = MAX(1, MIN(sdp->S, etos(1000., sdp->Lambda, sdp->K, sdp->H, Neff, sdp->efflen)));

		}
		else {
			if (S_set && !E_set)
				sdp->E = 0.;
			else
				sdp->E = E;
			sdp->S = S;
			cutoffs(&sdp->S, &sdp->E, &sdp->highs_exp, sdp->Lambda, sdp->K, sdp->H,
						Neff, sdp->efflen, 0.);
			Ecomp += sdp->E;
			/* Place our bets on how insignificant a hit will be obtained */
			if (sdp->E > 100.)
				sdp->S0 = MAX(1, MIN(sdp->S, etos(sdp->E*10., sdp->Lambda, sdp->K, sdp->H, Neff, sdp->efflen)));
			else
				sdp->S0 = MAX(1, MIN(sdp->S, etos(1000., sdp->Lambda, sdp->K, sdp->H, Neff, sdp->efflen)));
		}

		/* The following relation is safe to use because, should the score
		ever reach S during hit extension, the entire diagonal will be
		rigorously searched ignoring X.
		*/
		if (!X_set)
			sdp->X = ceil(10./(sdp->Lambda/LN2)); /* score for 10 bits */
		else
			sdp->X = X;
		sdp->X = MIN(sdp->X, sdp->S);
		sdp->minusX = -sdp->X;
	}
	if (i == maxframe - minframe)
		bfatal(ERR_SCORING, "The Karlin-Altschul parameters could not be determined in any reading frame");

	i=0;
	for (frame=minframe; frame<maxframe; ++frame) {
		sdp = &sd[frame];
		sdp->maxscore = fmaxscore(sdp->seq, sdp->len, maxcost);
		if (sdp->maxscore >= sdp->S)
			++i;
	}
	if (i == 0) {
		ckwarnings();
		(void) printf("\nThe maximum achievable scores are insufficient to satisfy the cutoff scores\n");
		(void) printf("\nThe max. achievable scores are:\n");
		(void) printf("        Frame  Max. Score\n");
		for (frame=minframe; frame<maxframe; ++frame)
			(void) printf("          %+d   %6d\n",
				sd[frame].stdframe, sd[frame].maxscore);
		(void) printf("\nYou might lower the cutoff score, raise the expect value, or use a different\nscoring scheme; or use a longer query sequence.\n");
		exit_code(ERR_SCORING);
	}
	Ecomp /= (maxframe - minframe);

	poolp = HSPPoolNew();
	hl_pool = HitListPoolNew();

	sys_times(&t1);

	/* build table of critical (neighborhood) words */
	for (dp = NULL, frame=minframe; frame<maxframe; ++frame) {
		sdp = &sd[frame];
		if (sdp->T <= 0)
			continue;
		if (!cdi_flag)
			dp = bld_table_aa(dp, frame, sdp->seq, sdp->len, W, sdp->T,
					alphasize, &matrx, &maxcost, &order, &numwords, &numexact);
		else
			dp = bld_table_xa(dp, frame, sdp->seq, aatab, W, sdp->T,
					alphasize, &matrx, &maxcost, &order, &numwords, &numexact);
	}
	end_table(dp);

	sys_times(&t2);

	RunWild("Searching", dbreccnt, dosearch);

#ifdef SIGXCPU
	/* Report results without interruption */
	signal(SIGXCPU, SIG_IGN);
#endif

	dfastatesize = dp->statesize;
	dfanstates = dp->nstates;
	dfasize = dfa_size(dp);
	dfaextent = dfa_extent(dp);
	dfa_destruct(dp);

#ifdef MPROC_AVAIL
	collate_hits();
	for (i=0; i<numprocs; ++i) {
		for (j=minframe; j<maxframe; ++j)
			sd[j].highs_obs = MAX(sd[j].highs_obs, mpc[i].sd[j].highs_obs);
	}
#endif

	hasHSP = hitlist_count(firsthl);
	sort_hitlists(&firsthl, hitlist_cmp_func);
	nheaders = MIN(hasHSP, MAX(V, showblast));
	if (nheaders < hasHSP)
		hitlist_truncate(&firsthl, nheaders);
	get_headers(DATABASE_SEQTYPE, firsthl);

	sys_times(&t3);
	print_headers(firsthl, 1.0, TRUE, 80);
	print_HSPs(firsthl, 1.0, print_x);
	print_parms();
	print_stats();
	ckwarnings();
	exit_code(ERR_NONE);
	/*NOTREACHED*/
}

/* print_parms() -- print the parameters */
print_parms()
{
	int		f, frame;
	int		sw=0, tw=0, xw=0;

	fflush(stdout);
	printf("\n\nParameters:\n");
	if (filtercmd != NULL && filtercmd[0] != NULLB)
		(void) printf("  Query filter:  \"%s\"\n", filtercmd);
	if (!E2_set && !S2_set)
		printf("  E = see table, S = %s,  E2 = n/a, S2 = n/a\n",
			(S_set ? ld2s((long)S) : "see table") );
	else
		printf("  E = %lg, S = %s,  E2 = see table, S2 = see table\n",
			Ecomp, (S_set ? ld2s((long)S) : "n/a") );
	printf("  W = %d, T = %s, X = %s\n",
		W,
		(T_set ? ld2s((long)T) : "see table"),
		(X_set ? ld2s((long)X) : "see table")
		);
	printf("  M = %s\n", M);
	printf("  C = %d (%s genetic code)\n", C, find_gcode(C)->name);
	printf("  H = n/a, V = %d, B = %d\n", V, showblast);
	if (cdi_flag)
		(void) printf("  -codoninfo %s\n", cdi_file);
	report_gapdecayrate();
	if (overlap_option != NULL)
		printf("  Overlap option:  %s\n", overlap_option);
	putchar('\n');

	for (f=minframe; f<maxframe; ++f) {
		sdp = &sd[f];
		sw = MAX(sw, longwidth10(sdp->S));
		tw = MAX(tw, longwidth10(sdp->T));
		xw = MAX(xw, longwidth10(sdp->X));
	}
	sw += 2, tw += 2, xw += 2;
	printf("  Frame  Length  Eff.Length    %s%*s%s%*sT%*sX\n",
		(E2_set || S2_set ? "E2" : " E"), sw-1, "", (E2_set || S2_set ? "S2" : " S"),
		tw-1, "", xw-1, "");
	for (f=minframe; f<maxframe; ++f) {
		frame = (f<3 ? 2-f : f); /* report in the order +3 +2 +1 -1 -2 -3 */
		sdp = &sd[frame];
		(void) printf("   %+d", sdp->stdframe);
		(void) printf("  %7s   %7s  %#8.2lg%*d%*d%*d\n",
				ld2s(sdp->len), ld2s(sdp->efflen), sdp->E,
				sw, sdp->S, tw, sdp->T, xw, sdp->X);
	}
	fflush(stdout);
	return ferror(stdout);
}

/* print_stats() -- print the statistics of the search */
print_stats()
{
	seq_data	*sdp;
	char	buf[20];
	int		f, frame, ew, fw, ow;

	fflush(stdout);
	ew = fw = ow = 0;
	for (f=minframe; f<maxframe; ++f) {
		sdp = &sd[f];
		ew = MAX(ew, longwidth10(sdp->highs_exp));
		ow = MAX(ow, longwidth10(sdp->highs_obs));
		sprintf(buf, "%.1lf", sdp->highs_exp*sdp->Lambda/LN2);
		fw = MAX(fw, (int)strlen(buf));
	}
	fw += 8;

	(void) printf("\nStatistics:%*sExpected  %*sObserved\n", 32, "", ew+fw-8, "");
	(void) printf("  Frame Lambda   Lambda/ln2   K       H    High Score%*s  High Score\n",
			ew+fw-10, "");

	printf("  Std%#8.3lg %#8.3lg    %#7.3lg %#7.3lg\n",
		Lambda_std, Lambda_std/LN2, K_std, H_std/LN2);
	for (f=minframe; f<maxframe; ++f) {
		frame = (f<3 ? 2-f : f); /* report in the order +3 +2 +1 -1 -2 -3 */
		sdp = &sd[frame];
		(void) printf(
			"   %+d%#8.3lg %#8.3lg    %#7.3lg %#7.3lg  %*ld (%.1lf bits)  %*ld (%.1lf bits) \n",
			sdp->stdframe,
			sdp->Lambda, sdp->Lambda/LN2, sdp->K, sdp->H/LN2,
			ew, (long)sdp->highs_exp, ((double)sdp->highs_exp)*sdp->Lambda/LN2,
			ow, (long)sdp->highs_obs, ((double)sdp->highs_obs)*sdp->Lambda/LN2
			);
		if (sdp->highs_exp > sdp->maxscore)
			(void) printf("  NOTE:  expected high score is greater than the max. achievable score %d\n",
					sdp->maxscore);
	}

	putchar('\n');
	(void) printf("  # of nucleotides in query:  %s", ld2s((long)n_len));
	if (Meff != n_len)
		(void) printf("  (Y = %s)\n", lu2s(Meff));
	else
		putchar('\n');
	(void) printf("  # of neighborhood words in query:  %s\n", lu2s((long)numwords));
	(void) printf("  # of exact words scoring below T:  %s\n", lu2s((long)numexact));
	(void) wrap(stdout, "  Database:  ", dbtitle, -1, 79, 13);
	(void) printf("  # of letters in database:  %s aa", lu2s((long)totdblen));
	if (Neff != totdblen)
		(void) printf("  (Z = %s)\n", lu2s(Neff));
	else
		putchar('\n');
	(void) printf("  # of word hits against database:  %s\n", lu2s((long)numhits));
	(void) printf("  # of failed hit extensions:  %s\n", lu2s((long)numfailed));
	(void) printf("  # of excluded hits:  %s\n",
				lu2s((long)numhits - (long)numfailed - totHSP - overcnt));
	(void) printf("  # of successful extensions:  %s\n", lu2s((long)totHSP+overcnt));
	(void) printf("  # of overlapping HSPs discarded:  %s\n",
				lu2s((long)overcnt));
	(void) printf("  # of HSPs reportable:  %s\n", lu2s((long)totHSP));
	(void) printf("  # of sequences in database:  %s\n", lu2s((long)nbr_seqs));
	(void) printf("  # of database sequences satisfying E:  %s\n",
				lu2s((long)hasHSP));

	(void) printf("No. of states in DFA:  %s (%s KB)\n", lu2s(dfanstates),
				lu2s(HOWMANY(dfanstates*dfastatesize, KBYTE)));
	(void) printf("Total size of DFA:  %s KB (%s KB)\n",
				lu2s(HOWMANY(dfasize, KBYTE)),
				lu2s(HOWMANY(dfaextent, KBYTE)));
	sys_showtime("Time to generate neighborhoods", &t1, &t2, stdout);
#ifdef MPROC_AVAIL
	(void) printf("No. of processors used:  %d\n", numprocs);
#endif
	sys_showtime("Time to search database", &t2, &t3, stdout);
	sys_times(&t3);
	sys_showtime("Total cpu time", (Cputime_t *)NULL, &t3, stdout);
	fflush(stdout);
	return ferror(stdout);
}

unsigned	hashsp;

static void
dosearch(userp, tp)
	Nlm_VoidPtr	userp;
	TaskBlkPtr	tp;
{
	int		hashsp_local = 0, rem_max_per_proc;
	Boolean	saved = FALSE;
	register HSPPtr	hp;
	MPDEF

	rem_max_per_proc = REMEMBER_MAX / tp->proc_max;
#ifdef MPROC_AVAIL
	initmpc(tp->id, mp = &mpc[tp->id]);
#endif
	/* For each sequence in the database */
	while (TaskNext(tp) >= 0) {
		get_seq(MPC tp->task_cur + dbrecmin);

		/* search the database sequence */
		search_x(MP);

		/* Save any HSPs that were found */
		if (MPPTR curhl.hp != NULL) {
			sort_HSPs(&MPPTR curhl);
			if (consistency_flag)
				consistn(&MPPTR curhl);
			else
				pcnt(&MPPTR curhl);
			if (S2_set || E2_set) {
				if (consistency_flag)
					consist_evals_x(&MPPTR curhl, MPPTR s_len, Neff);
				else
					evals_x(&MPPTR curhl, MPPTR s_len, Neff);
				hitlist_prune_by_cutoff(&MPPTR curhl, MPPTR poolp, SCORET_EXPECT, Ecomp);
				if (MPPTR curhl.hp == NULL)
					continue;
				for (hp = MPPTR curhl.hp; hp != NULL; hp = hp->next)
					hp->P_val = etop(hp->expect);
			}
			else {
				if (consistency_flag)
					consist_pvals_x(&MPPTR curhl, MPPTR s_len, Neff);
				else
					pvals_x(&MPPTR curhl, MPPTR s_len, Neff);
			}
			MPPTR curhl.seq_nbr = tp->task_cur + dbrecmin;
			if (++hashsp_local + hashsp > rem_max_per_proc) {
				M_LOCK;
				if (!saved) {
					saved = TRUE;
					hashsp += hashsp_local - 1;
					hashsp_local = 0;
				}
				if (++hashsp > REMEMBER_MAX)
					bfatal(ERR_REMEMBER,
					"More than %u sequences have HSPs to remember; increase REMEMBER_MAX and recompile.",
					REMEMBER_MAX);
				M_UNLOCK;
			}
			hitlist_save(MPPTR hl_pool, &MPPTR firsthl, &MPPTR curhl);
			MPPTR totHSP += MPPTR curhl.hspcnt;
			MPPTR curhl.hp = NULL;
			MPPTR curhl.hspcnt = 0;
		}
	}
	M_LOCK;
	hashsp += hashsp_local;
	M_UNLOCK;
}

#ifdef MPROC_AVAIL
initmpc(id, mp)
	int		id;
	MPContextPtr	mp;
{
	int		i, frame, len;
	seq_data	*sdp;

	mp->id = id;
	MemCpy((char *)&mp->sd[0], (char *)&sd[0], sizeof(sd));
	if (id == 0) {
		mp->poolp = poolp;
		mp->hl_pool = hl_pool;
	}
	else {
		mp->poolp = HSPPoolNew();
		mp->hl_pool = HitListPoolNew();
	}
	for (frame=minframe; frame<maxframe; ++frame) {
		sdp = &mp->sd[frame];
		len = maxlen + sdp->len + 1;
		sdp->_dlev = (DIAG_TYPE *)ckalloc0(sizeof(sdp->_dlev[0])*len);
		sdp->diag_top = sdp->diag_stack =
			(DIAG_PTR)ckalloc(sizeof(sdp->diag_top[0])*len);
		sdp->diag_level = sdp->_dlev + sdp->len;
	}
}
#endif


display_gcodes()
{                
    register int    i;
                       
    printf("The available genetic codes are:\n");
    for (i=0; i<DIM(gcodes); ++i)                
        printf("  %d. %s\n", gcodes[i].id, gcodes[i].name);
    printf("Specify a particular code by its number (%d is the default)\n",
		DEFAULT_C);
}

/* get_query - read the query sequence */
get_query(filename)
	char	*filename;
{
	FILE	*fp;
	char	*base, *bp;
	size_t	basemax;
	register char *p, ch;
		
	fp = ckopen(filename, "r", 1);
	base = bp = (char *)ckalloc(basemax = 5*KBYTE);

	while (vfgets(&base, &basemax, &bp, fp) != NULL) {
		if (*bp == '>')	{ /* Optional header/title line */
			if (bp != base) /* Second sequence encountered */
				break;
			p = str_chr(bp, '\n');
			if (p != NULL)
				*p = NULLB;
			qtitle = str_dup(bp+1);
			qtitlelen = strlen(qtitle);
			for (p=qtitle; (ch = *p) && !isspace(ch); ++p)
				;
			*p = NULLB;
			qname = str_dup(qtitle);
			*p = ch;
			qnamelen = strlen(qname);
			continue;
		}
		p = bp;
		while ((ch = *p++) != NULLB) {
			if (nt_atob[ch] <= NUCID_MAX) {
				if (isalpha(ch))
					if (islower(ch))
						ch = toupper(ch);
				*bp++ = ch;
			}
			else
				if (nt_atob[ch] == NUCID_NAR) {
					nt_atob[ch] = NUCID_IGNORE;
warning("Invalid nucleotide code ``%c'' encountered in query sequence will be skipped.", ch);
				}
		}
		if (bp - base >= QUERYLEN_MAX)
			bfatal(ERR_QUERYLEN, "query sequence exceeds %d characters in length",
				QUERYLEN_MAX-1);
			
	}
	(void) fclose(fp);
	n_seq = base;
	n_len = bp - n_seq;
	*bp = NULLB;
	qq_seq = (char *)ckalloc(n_len/3+1);

	(void) ckseqtype(QUERY_SEQTYPE, n_seq, n_len);
	return 0;
}

/* nucleotide sequence --> amino acid sequence translation routines */

xlat_query()
{
	char	*revcomp();
	int		i, len;
	char	name[30];

	for (frame=minframe; frame<3; ++frame) {
		sdp = &sd[frame];
		sprintf(name, "Frame%+d", sdp->stdframe);
		i = strlen(name);
		xlat_seq(n_seq+frame, n_len-frame, &sdp->seq, &sdp->len);
		if (seqfilter(filtercmd, aa_btoa, aa_atob,
					name, i,
					(UcharPtr)sdp->seq, sdp->len, (UcharPtr)sdp->seq,
					(size_t PNTR)&sdp->len) != 0) {
			bfatal(ERR_UNDEF, "Non-zero return code from sequence filter");
		}
	}
	if (bottom)
		n_rev = revcomp(n_seq, n_len);
	for (frame=3; frame<maxframe; ++frame) {
		sdp = &sd[frame];
		sprintf(name, "Frame%+d", sdp->stdframe);
		i = strlen(name);
		xlat_seq(n_rev+frame-3, n_len-frame+3, &sdp->seq, &sdp->len);
		if (seqfilter(filtercmd, aa_btoa, aa_atob,
				name, i,
				(UcharPtr)sdp->seq, sdp->len, (UcharPtr)sdp->seq,
				(size_t PNTR)&sdp->len) != 0) {
			bfatal(ERR_UNDEF, "Non-zero return code from sequence filter");
		}
	}
}

init_diags()
{
	int	frame, len;

	for (frame=minframe; frame<maxframe; ++frame) {
		sdp = &sd[frame];
		len = maxlen + sdp->len + 1;
		sdp->_dlev = (DIAG_TYPE *)ckalloc0(sizeof(sdp->_dlev[0])*len);
		sdp->diag_top = sdp->diag_stack =
					(DIAG_PTR)ckalloc(sizeof(sdp->diag_top[0])*len);
		sdp->diag_level = sdp->_dlev + sdp->len;
	}
}

xlat_seq(nseq, nlen, aseq, alen)
	register char	*nseq;
	int		nlen;
	char	**aseq;
	int		*alen;
{
	register char	*cp;
	register int	len = nlen/3;
	register unsigned char	aa;
	unsigned n1, n2, n3;

	if (len < 0)
		len = 0;

	cp = (char *)ckalloc(len + 2);
	*aseq = (cp += 1);
	cp[-1] = NULLB; /* null prefix byte */

	while (nlen > 2) {
		nlen -= CODON_LEN;
		n1 = nt_atob[nseq[0]];
		n2 = nt_atob[nseq[1]];
		n3 = nt_atob[nseq[2]];
		if (!cdi_flag)
			aa = codon2aa(aatab, n1, n2, n3);
		else {
			if (n1 < 4 && n2 < 4 && n3 < 4)
				aa = n1*(4*4) + n2*4 + n3 + 1;
			else
				aa = (64+1);
		}
		nseq += CODON_LEN;
		*cp++ = aa;
	}
	*cp = NULLB;
	*alen = cp - *aseq;
}


char *
revcomp(s, len)
	register char	*s;
	register int	len;
{
	register char	*t;

	t = (char *)ckalloc(len+1);
	t[len] = NULLB;
	while (len-- > 0)
		t[len] = nt_arevcomp[*s++];
	return t;
}

/* UNIX-specific input routine; database must be pre-processed by SETDB */

static void
initdb(dbname)
	char	*dbname;
{
	struct stat	sbuf;
	long	dbtype, dbformat, i;
	char	*dirname = "";
	char	name[FILENAME_MAX];
	long	many, len;
	int		once=0;

RETRY:
	if (*dirname != NULLB)
		(void) sprintf(dbfname, "%s%s%s", dirname, DIRDELIMSTR, dbname);
	else
		SAFENCPY(dbfname, dbname, sizeof(dbfname));

	(void) sprintf(name, "%s%s", dbfname, TABLE_EXT);
	tfile = mfil_open(name, "r", MFIL_OPT_ALL);
	if (tfile == NULL) {
		if ((dirname = getenv("BLASTDB")) == NULL)
			dirname = BLASTDB;
		if (++once == 1)
			goto RETRY;
		bfatal(ERR_FOPEN, "Couldn't open database file ``%s''.  Make sure you have specified the correct database name and requested a database of the correct type (protein) for use with this program.", basename(name, NULL));
	}

	if (mfil_dup_long(tfile, &dbtype, 1) != 1)
		bfatal(ERR_DBASE, "Database read error on file %s", name);
	if (dbtype != DB_TYPE_PRO)
		bfatal(ERR_DBASE, "``%s'' is not a protein sequence database.", dbname);
	if (mfil_dup_long(tfile, &dbformat, 1) != 1)
		bfatal(ERR_DBASE, "Database read error on file %s", name);
	if (dbformat != AAFORMAT)
		bfatal(ERR_DBASE, "Database format is an incompatible version; re-run ``setdb''.");

	if (mfil_dup_long(tfile, &i, 1) != 1)
		bfatal(ERR_DBASE, "Corrupted database file:  %s", name);
	dbtitle = mfil_get(tfile, i);
	if (i%4 != 0) {                                      
		i = 4 - (i%4);
		mfil_seek(tfile, i, SEEK_CUR);
	}

	(void) mfil_dup_long(tfile, (long *)&nbr_seqs, 1);
	(void) mfil_dup_long(tfile, (long *)&totdblen, 1);
	if (mfil_dup_long(tfile, (long *)&maxlen, 1) != 1)
		bfatal(ERR_DBASE, "corrupted database file:  %s", name);

	if (dbtitle[0] == NULLB)
		dbtitle = str_dup(dbname);
	(void) wrap(stdout, "Database:  ", dbtitle, -1, 79, 11);
	(void) printf("           %s sequences; %s total letters.\n",
		ld2s(nbr_seqs), ld2s(totdblen));                       

	if (nbr_seqs <= 0 || totdblen <= 0)
		bfatal(ERR_DBASE, "Nothing in the database to search!?");

	many = nbr_seqs + 1;
	ss_seq = (char *)ckalloc(sizeof(*ss_seq)*(maxlen+1));

	seq_beg = (unsigned long *)mfil_get_long(tfile, many);
	header_beg = (unsigned long *)mfil_get_long(tfile, many);
	if (seq_beg == NULL || header_beg == NULL)
		bfatal(ERR_DBASE, "improper offsets in database table file");

	(void) sprintf(name, "%s%s", dbfname, AA_HEADER_EXT);
	hfile = mfil_open(name, "r", MFIL_OPT_ALL);
	if (hfile == NULL)
		bfatal(ERR_FOPEN, "Couldn't open database file %s", name);

	(void) sprintf(name, "%s%s", dbfname, AA_SEARCHSEQ_EXT);
	/* Attach a shared memory segment for the .bsq file (if one exists) */
	sfile = mfil_open(name, "r", MFIL_OPT_ALL);
	if (sfile == NULL)
		bfatal(ERR_FOPEN, "Couldn't open database file %s", name);
	if ((dbbuf = MFILE_SP(sfile)) == NULL) {
#ifdef MPROC_AVAIL
		off_t	fsize;
		fflush(stdout);
		fstat(fileno(MFILE_FP(sfile)), &sbuf);
		fsize = sbuf.st_size;
		dbbuf = ckalloc(fsize);
		if (mfil_read(dbbuf, fsize, 1, sfile) != 1)
			bfatal(ERR_DBASE, "Error reading database file %s", name);
#else
		(void) mfil_seek(sfile, (long)seq_beg[0], SEEK_SET);
		s_seq = (char *)ckalloc(sizeof(*s_seq)*(maxlen+2));
		*s_seq++ = NULLB;
#endif
	}

	len = maxlen + q_len + 1;

	if (Neff == 0)
		Neff = totdblen;

	dbrecmin = MAX(dbrecmin, 0);
	dbrecmin = MIN(dbrecmin, nbr_seqs);
	dbrecmax = (dbrecmax > 0 ? MIN(dbrecmax, nbr_seqs) : nbr_seqs);
	dbreccnt = dbrecmax - dbrecmin;
	if (dbreccnt < nbr_seqs)
		printf("           Subset of database from %ld to %ld.\n",
				dbrecmin+1, dbrecmax);
}

/* get_seq - read a database sequence and header line, return 0 on EOF */
static int
get_seq(MPC current)
	MPDEF
	unsigned long	current;
{
	if (dbbuf != NULL) {
		MPPTR s_len = seq_beg[current+1] - seq_beg[current] - 1;
		MPPTR s_seq = dbbuf + seq_beg[current];
	}
	else {
		s_len = seq_beg[current+1] - seq_beg[current];
		if (mfil_read(s_seq, 1, s_len, sfile) != s_len)
			bfatal(ERR_DBASE, "error reading database sequence.");
		--s_len;
	}
	if (MPPTR s_len > maxlen)
		bfatal(ERR_DBASE, "Database sequence longer than expected maxlen!");
	return 0;
}

/***************************/
#include "searchx.inc"
/***************************/

/* ----------------------------  parameter calculations  ---------------------*/
/* Version of lambdak() specifically for a.a. alphabet used by BLASTP */
int
lambdak(lows, highs, q_seq, q_len, efflen, Lambda, K, H)
	int		lows, highs, q_len;
	long	PNTR efflen;
	register CharPtr	q_seq;
	double	PNTR Lambda, PNTR K, PNTR H;
{
	double	fq1[DIM(matrx[0])];
	long	iq1[DIM(matrx[0])];
	double	*prob;
	unsigned long	qlen;
	register int	i, j;

	if (lows >= 0 || highs <= 0)
		return 1;

	/* Tally the frequency of occurrence of each letter in the query */
	for (i=0; i<DIM(matrx[0]); ++i)
		iq1[i]=0;
	for (i=0, qlen = q_len; i<q_len; ++i) {
		if (q_seq[i] != aa_atob[UNKNOWN_AA_CHR])
			++iq1[q_seq[i]];
		else
			--qlen;
	}
	*efflen = qlen;

	/* Convert frequencies to floating point */
	for (i=0; i<DIM(fq1); ++i)
		fq1[i] = iq1[i];

	/* Convert frequencies to probabilities */
	for (i=0; i<DIM(fq1); ++i)
		fq1[i] /= qlen;

	/* Calculate probability of each score */
	prob = (double *)ckalloc((highs-lows+1) * sizeof(*prob));
	for (i=highs-lows; i>=0; --i)
		prob[i]=0.;
	for (i=minchar; i<=maxchar; ++i)
		for (j=1; j<DIM(aafq); ++j) {
			if (matrx[i][j] > SCORE_MIN/WORDSIZE_MAX)
				prob[matrx[i][j]-lows] += fq1[i]*aafq[j];
		}

	i = karlin(lows, highs, prob, !L_set, Lambda, !K_set, K, TRUE, H);
	mem_free((char *)prob);
	if (i != 0) {
		*K = *Lambda = -1.;
		return i;
	}
	return 0;
}

/* lambdak_std -- compute L,K,H for "average" protein query vs. average db */
int
lambdak_std(lows, highs, Lambda, K, H)
	int		lows, highs;
	double	PNTR Lambda, PNTR K, PNTR H;
{
	double	fq1[DIM(matrx[0])];
	long	iq1[DIM(matrx[0])];
	double	*prob;
	unsigned long	qlen;
	register int	i, j;

	/* Calculate probability of each score */
	prob = (double *)ckalloc((highs-lows+1) * sizeof(*prob));
	for (i=highs-lows; i>=0; --i)
		prob[i]=0.;
	for (i=1; i<DIM(aafq); ++i)
		for (j=1; j<DIM(aafq); ++j) {
			if (matrx[i][j] > SCORE_MIN/WORDSIZE_MAX)
				prob[matrx[i][j]-lows] += aafq[i]*aafq[j];
		}

	i = karlin(lows, highs, prob, !L_set, Lambda, !K_set, K, TRUE, H);
	mem_free((char *)prob);
	if (i != 0) {
		*K = *Lambda = -1.;
		return i;
	}
	return 0;
}

/* Version of lambdak() specifically for a.a. alphabet used by BLAST-Codons */
long	iq1[DIM(matrx[0])];

lambdak0()
{
	register int	i;

	for (i = 0; i < DIM(iq1); ++i)
		iq1[i] = 0;
	return 0;
}

lambdakcnt(q_seq, q_len)
	register CharPtr	q_seq;
	register int	q_len;
{
	register int	i;

	/* Tally the frequency of occurrence of each letter in the query */
	for (i = 0; i < q_len; ++i)
		++iq1[nt_atob[q_seq[i]]];
	return 0;
}

int
lambdakcodon(lows, highs, efflen, Lambda, K, H)
	int	lows, highs;
	long	PNTR efflen;
	double	*Lambda, *K, *H;
{
	double	fq1[4];
	double	*prob, p1, p2, p3;
	register int	i, j, tot;

	if (lows >= 0 || highs <= 0)
		return 1;

	/* convert frequencies to floating point */
	for (i = 0, tot = 0; i < DIM(fq1); ++i) {
		tot += iq1[i];
		fq1[i] = iq1[i];
	}
	if (tot == 0)
		bfatal(ERR_UNDEF, "None of the residues in the query sequence are A, C, G, or T");
	if (Y_set)
		*efflen = Meff/CODON_LEN;
	else
		*efflen = tot/CODON_LEN;

	/* convert freqs. to probabilities */
	for (i = 0; i < DIM(fq1); ++i)
		fq1[i] /= tot;

	/* calculate probability of occurance of each score */
	prob = (double *)ckalloc0((highs - lows + 1) * sizeof(*prob));
	for (i = 1; i < 65; ++i) {
		p1 = fq1[(i-1) & 3];
		p2 = fq1[((i-1)>>2) & 3];
		p3 = fq1[((i-1)>>4) & 3];
		for (j = 1; j < DIM(aafq); ++j) {
			if (matrx[i][j] >= lows)
				prob[matrx[i][j] - lows] += p1*p2*p3*aafq[j];
		}
	}

	i = karlin(lows, highs, prob, !L_set, Lambda, !K_set, K, TRUE, H);
	mem_free((CharPtr)prob);
	if (i != 0) {
		*Lambda = *K = *H = -1.;
		return i;
	}
	return 0;
}

static int
print_x(hlp)
	HitListPtr	hlp;
{
	Seqid_t	current;
	int		i, len, prevframe = 0;
	HSPPtr	hp;
	char	*btoa;
	register char	*cp, ch;

	current = hlp->seq_nbr;

	cp = hlp->header;
	while ((ch = *cp) != NULLB && !isspace(ch))
		++cp;
	while ((ch = *cp) != NULLB && isspace(ch))
		++cp;
	i = cp - hlp->header + 1;
	i = MIN(i, 12);
	putchar('\n');
	putchar('\n');
	(void) wrap(stdout, ">", hlp->header, hlp->header_len, 79, i);

	/* get sequence */
	if (dbbuf != NULL) {
		len = seq_beg[current+1] - seq_beg[current] - 1;
		s_seq = dbbuf + seq_beg[current];
	}
	else {
		len = seq_beg[current+1] - seq_beg[current];
		if (mfil_seek(sfile, seq_beg[current], SEEK_SET) == -1)
			bfatal(ERR_DBASE, "fseek error on database sequence file");
		if (mfil_read(s_seq, 1, len, sfile) != len)
			bfatal(ERR_DBASE, "error reading database sequence");
		--len;
	}
	(void) printf("%*sLength = %s\n", i, "", ld2s((long)len));

	if (!cdi_flag)
		btoa = aa_btoa;
	else
		btoa = xa_btoa;
	
	for (hp = hlp->hp; hp != NULL; hp = hp->next) {
		frame = hp->frame;
		sdp = &sd[(frame > 0 ? frame-1 : 2-frame)];
		q_seq = sdp->seq;

		if (SIGN(prevframe) != SIGN(frame)) {
			if (frame > 0)
				printf("\n  Plus");
			else
				printf("\n  Minus");
			printf(" Strand HSPs:\n");
		}

		if (hsp_print(stdout, hp, "", (UcharPtr)q_seq + hp->q_pos, (UcharPtr)s_seq + hp->s_pos, matrx, btoa, aa_btoa, Qlen, s_len, apo_to_nt1, apo_to_nt3, nox_coord, nox_coord, "Frame", str_frame, sdp->Lambda) != 0)
			return -1;
		prevframe = frame;
	}
	return 0;
}

usage()
{
	fflush(stdout);
	fflush(stderr);
	fprintf(stderr, "\nUsage:\n\n");
	fprintf(stderr,
		"    %s database queryfile [options]\n\n", module);
	fprintf(stderr,
		"Valid %s options:  E, S, W, T, X, M, C, Y, Z, K, L, V, and B\n",
		module);
	fprintf(stderr, "\t-qoffset #\n");
	fprintf(stderr, "\t-dbrecmin #\n");
	fprintf(stderr, "\t-dbrecmax #\n");
	fprintf(stderr, "\t-gi (gi identifier suppression)\n");
	fprintf(stderr, "\ttop\n\tbottom\n\t-overlap\n\t-overlap2\n\t-gapdecayrate #\n\t-filter filteroption\n\t-echofilter\n\t-consistency\n");
	fprintf(stderr, "\t-codoninfo cdifile\n");
	fprintf(stderr, "\t-sort_by_pvalue\n\t-sort_by_count\n\t-sort_by_highscore\n\t-sort_by_totalscore\n");
	exit_code(ERR_INVAL);
}
