/**************** Statistical Significance Parameter Subroutine ****************

	Version 1.0	February 2, 1990

	Program by:	Stephen Altschul

	Address:	National Center for Biotechnology Information
			National Library of Medicine
			National Institutes of Health
			Bethesda, MD  20894

	Internet:	altschul@ncbi.nlm.nih.gov

See:	Karlin, S. & Altschul, S.F. "Methods for Assessing the Statistical
	Significance of Molecular Sequence Features by Using General Scoring
	Schemes,"  Proc. Natl. Acad. Sci. USA 87 (1990), 2264-2268.

	Computes the parameters lambda and K for use in calculating the
	statistical significance of high-scoring segments or subalignments.

	The scoring scheme must be integer valued.  A positive score must be
	possible, but the expected (mean) score must be negative.

	A program that calls this routine must provide the value of the lowest
	possible score, the value of the greatest possible score, and a pointer
	to an array of probabilities for the occurence of all scores between
	these two extreme scores.  For example, if score -2 occurs with
	probability 0.7, score 0 occurs with probability 0.1, and score 3
	occurs with probability 0.2, then the subroutine must be called with
	low = -2, high = 3, and pr pointing to the array of values
	{ 0.7, 0.0, 0.1, 0.0, 0.0, 0.2 }.  The calling program must also provide
	pointers to lambda and K; the subroutine will then calculate the values
	of these two parameters.  In this example, lambda=0.330 and K=0.154.

	The parameters lambda and K can be used as follows.  Suppose we are
	given a length N random sequence of independent letters.  Associated
	with each letter is a score, and the probabilities of the letters
	determine the probability for each score.  Let S be the aggregate score
	of the highest scoring contiguous segment of this sequence.  Then if N
	is sufficiently large (greater than 100), the following bound on the
	probability that S is greater than or equal to x applies:
	
		P( S >= x )   <=   1 - exp [ - KN exp ( - lambda * x ) ].
	
	In other words, the p-value for this segment can be written as
	1-exp[-KN*exp(-lambda*S)].

	This formula can be applied to pairwise sequence comparison by assigning
	scores to pairs of letters (e.g. amino acids), and by replacing N in the
	formula with N*M, where N and M are the lengths of the two sequences
	being compared.

	In addition, letting y = KN*exp(-lambda*S), the p-value for finding m
	distinct segments all with score >= S is given by:

                               2             m-1           -y
                1 - [ 1 + y + y /2! + ... + y   /(m-1)! ] e

	Notice that for m=1 this formula reduces to 1-exp(-y), which is the same
	as the previous formula.

*******************************************************************************/
#include <ncbi.h>
#include <gishlib.h>
#include "error.h"
#include "karlin.h"

#define MAXIT 20	/* Maximum number of iterations used in calculating K */

/* For greater accuracy, set SUMLIMIT to 0.00001 */
/* For higher speed, set SUMLIMIT to 0.001 */
#define SUMLIMIT 0.01

int
karlin(low, high, pr, dol, lambda, dok, K, doh, H)
	Score_t	low;		/* Lowest score (must be negative)    */
	Score_t	high;		/* Highest score (must be positive)   */
	double	*pr;		/* Probabilities for various scores   */
	int		dol;
	double	*lambda;	/* Pointer to parameter lambda        */
	int		dok;
	double	*K;			/* Pointer to parmeter K              */
	int		doh;
	double	*H;			/* Pointer to parmeter K              */
{
	int		i, j;
	Score_t	range, lo, hi, first, last;
	double	up, new, sum, Sum, av, beta, oldsum, oldsum2;
	double	*p = NULL, *P = NULL, *ptrP, *ptr1, *ptr2;
	double	ratio;

	/* Check that scores and their associated probabilities are valid     */

	if (low >= 0.)
		bfatal(ERR_SCORING, "There must be at least one negative score in the substitution matrix.");
	for (i=range=high-low; i> -low && pr[i] == 0.; --i)
		;
	if (i <= -low)
		bfatal(ERR_SCORING,
		"A positive score is impossible in the context of the scoring scheme, the residue composition of the query sequence, and the residue composition assumed for the database.");
	for (sum=i=0; i<=range; sum += pr[i++])
		if (pr[i] < 0.)
			bfatal(ERR_SCORING, "Negative probabilities for scores are disallowed.");
	if (sum<0.99995 || sum>1.00005)
		warning("Score probabilities sum to %.5lf and will be normalized to 1.", sum);
	p = (double *)ckalloc0(sizeof(*p) * (range+1));
	for (Sum=low,i=0; i<=range; ++i)
		Sum += i*(p[i]=pr[i]/sum);
	if (Sum >= 0.)
		bfatal(ERR_SCORING,
			"Invalid (non-negative) expected score:  %#0.3lg", Sum);

	/* Calculate the parameter lambda */

	if (dol) {
		up = 0.5;
		do {
			up *= 2;
			ptr1 = p;
			for (sum=0,i=low; i<=high; ++i)
				sum += *ptr1++ * exp(up*i);
		} while (sum<1.0);

		/* Root solving by the bisection method */
		for (*lambda=0., j=0; j<25; ++j) {
			new = (*lambda+up)/2.0;
			ptr1 = p;
			for (sum=0., i=low; i <= high; ++i)
				sum += *ptr1++ * exp(new*i);
			if (sum > 1.0)
				up = new;
			else
				*lambda = new;
		}
	}
	beta = exp(*lambda);

	/* Calculate the relative entropy of the p's and q's, parameter H */
	ptr1 = p;
	for (av=0, i=low; i<=high; ++i)
		av += *ptr1++ *i*exp(*lambda*i);
	if (doh)
	    *H = *lambda*av;

	/* Calculate the parameter K */
	if (!dok)
		goto OKExit;
	if (low == -1 || high == 1) {
		*K = (high == 1 ? av : Sum*Sum/av);
		*K *= 1.0 - 1./beta;
		goto OKExit;
	}
	Sum = 0.;
	lo = hi = 0;
	P = (double *)ckalloc0(MAXIT* (range+1) * sizeof(*P));
	*P = sum = oldsum = oldsum2 = 1.;
	for (j=0;  j<MAXIT && sum > SUMLIMIT; oldsum = sum, Sum += sum /= ++j) {
		first = last = range;
		for (ptrP = P + (hi += high) - (lo += low); ptrP >= P; *ptrP-- =sum) {
			ptr1 = ptrP - first;
			ptr2 = p + first;
			for (sum=0., i=first; i <= last; ++i)
				sum += *ptr1-- * *ptr2++;
			if (first != 0)
				--first;
			if (ptrP-P <= range)
				--last;
		}
		new = fct_powi(beta, lo-1);
		for (sum=0, i=lo; i != 0; ++i)
			sum += *++ptrP * (new *= beta);
		for (; i <= hi; ++i)
			sum += *++ptrP;
		oldsum2 = oldsum;
	}

	/* Geometric progression correction terms to accommodate fewer iterations */
	ratio = oldsum / oldsum2;
	if (ratio >= (1.0 - SUMLIMIT*0.001))
		bfatal(ERR_SCORING, "Value calculated for K was too high due to insufficient iterations.  Alternatively, the expected average score is insufficiently negative.");
	while (sum > SUMLIMIT*0.01) {
		oldsum *= ratio;
		Sum += sum = oldsum / ++j;
	}

	for (i=low; p[i-low] == 0.; ++i)
		;
	for (j= -i;i<high && j>1;)
		if (p[++i-low])
			j = fct_gcd(j,i);

	*K = (j*exp(-2.*Sum))/(av*etop(*lambda * j));

OKExit:
	if (p != NULL)
		mem_free(p);
	if (P != NULL)
		mem_free(P);
	return 0;		/* Parameters calculated successfully */
}
